#!/usr/bin/env python
#  SSSD
#
#  SSSD Config API tests
#
#  Copyright (C) Red Hat
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Created on Sep 18, 2009

@author: sgallagh
"""
import unittest
import os
import shutil
import tempfile
from stat import ST_MODE, S_IMODE

import sys

srcdir = os.getenv('srcdir')
if srcdir:
    sys.path.insert(0, "./src/config")
    srcdir = srcdir + "/src/config"
else:
    srcdir = "."
import SSSDConfig  # noqa


def create_temp_dir():
    test_dir = os.environ.get('SSS_TEST_DIR') or "."
    return tempfile.mkdtemp(dir=test_dir)


def striplist(the_list):
    return ([x.strip() for x in the_list])


class SSSDConfigTestValid(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = create_temp_dir()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def testServices(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Validate services
        services = sssdconfig.list_services()
        self.assertTrue('sssd' in services)
        self.assertTrue('nss' in services)
        self.assertTrue('pam' in services)

        # Verify service attributes
        sssd_service = sssdconfig.get_service('sssd')
        service_opts = sssd_service.list_options()

        self.assertTrue('services' in service_opts.keys())
        service_list = sssd_service.get_option('services')
        self.assertTrue('nss' in service_list)
        self.assertTrue('pam' in service_list)

        self.assertTrue('domains' in service_opts)

        del sssdconfig
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.new_config()
        sssdconfig.delete_service('sssd')
        new_sssd_service = sssdconfig.new_service('sssd')
        new_options = new_sssd_service.list_options()

        self.assertTrue('debug_level' in new_options)
        self.assertEqual(new_options['debug_level'][0], int)

        self.assertTrue('command' in new_options)
        self.assertEqual(new_options['command'][0], str)

        self.assertTrue('services' in new_options)
        self.assertEqual(new_options['debug_level'][0], int)

        self.assertTrue('domains' in new_options)
        self.assertEqual(new_options['domains'][0], list)
        self.assertEqual(new_options['domains'][1], str)

        self.assertTrue('re_expression' in new_options)
        self.assertEqual(new_options['re_expression'][0], str)

        self.assertTrue('full_name_format' in new_options)
        self.assertEqual(new_options['full_name_format'][0], str)

        self.assertTrue('default_domain_suffix' in new_options)
        self.assertEqual(new_options['default_domain_suffix'][0], str)

        self.assertTrue('domain_resolution_order' in new_options)
        self.assertEqual(new_options['domain_resolution_order'][0], list)
        self.assertEqual(new_options['domain_resolution_order'][1], str)

        del sssdconfig

    def testDomains(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Validate domain list
        domains = sssdconfig.list_domains()
        self.assertTrue('LDAP' in domains)
        self.assertTrue('PROXY' in domains)
        self.assertTrue('IPA' in domains)

        # Verify domain attributes
        ipa_domain = sssdconfig.get_domain('IPA')
        domain_opts = ipa_domain.list_options()
        self.assertTrue('debug_level' in domain_opts.keys())
        self.assertTrue('id_provider' in domain_opts.keys())
        self.assertTrue('auth_provider' in domain_opts.keys())
        self.assertEqual(ipa_domain.get_option('debug_level'), 0xff0)

        proxy_domain = sssdconfig.get_domain('PROXY')
        self.assertEqual(proxy_domain.get_option('debug_level'), 1)

        # Verify attributes in responders
        pam_responder = sssdconfig.get_service('pam')
        self.assertEqual(pam_responder.get_option('debug_level'), 2)

        sudo_responder = sssdconfig.get_service('sudo')
        self.assertEqual(sudo_responder.get_option('debug_level'), 0xfc10)

        del sssdconfig

    def testListProviders(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        sssdconfig.new_config()
        junk_domain = sssdconfig.new_domain('junk')
        providers = junk_domain.list_providers()
        self.assertTrue('ldap' in providers.keys())

    def testCreateNewLDAPConfig(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        sssdconfig.new_config()

        ldap_domain = sssdconfig.new_domain('LDAP')
        ldap_domain.add_provider('ldap', 'id')
        ldap_domain.set_option('debug_level', 1)
        ldap_domain.set_active(True)
        sssdconfig.save_domain(ldap_domain)

        of = self.tmp_dir + '/testCreateNewLDAPConfig.conf'

        # Ensure the output file doesn't exist
        try:
            os.unlink(of)
        except OSError:
            pass

        # Write out the file
        sssdconfig.write(of)

        # Verify that the output file has the correct permissions
        mode = os.stat(of)[ST_MODE]

        # Output files should not be readable or writable by
        # others, and should not be executable by anyone
        self.assertFalse(S_IMODE(mode) & 0o137)

        # try to import saved configuration file
        config = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                       srcdir + "/etc/sssd.api.d")
        config.import_config(configfile=of)

        # Remove the output file
        os.unlink(of)

    def testModifyExistingConfig(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        ldap_domain = sssdconfig.get_domain('LDAP')
        ldap_domain.set_option('debug_level', 3)

        ldap_domain.remove_provider('auth')
        ldap_domain.add_provider('krb5', 'auth')
        ldap_domain.set_active(True)
        sssdconfig.save_domain(ldap_domain)

        proxy_domain = sssdconfig.get_domain('PROXY')
        proxy_domain.set_option('debug_level', 0x1f10)
        sssdconfig.save_domain(proxy_domain)

        sudo_responder = sssdconfig.get_service('sudo')
        sudo_responder.set_option('debug_level', 0x2210)
        sssdconfig.save_service(sudo_responder)

        pam_responder = sssdconfig.get_service('pam')
        pam_responder.set_option('debug_level', 9)
        sssdconfig.save_service(pam_responder)

        of = self.tmp_dir + '/testModifyExistingConfig.conf'

        # Ensure the output file doesn't exist
        try:
            os.unlink(of)
        except OSError:
            pass

        # Write out the file
        sssdconfig.write(of)

        # Verify that the output file has the correct permissions
        mode = os.stat(of)[ST_MODE]

        # Output files should not be readable or writable by
        # others, and should not be executable by anyone
        self.assertFalse(S_IMODE(mode) & 0o137)

        # try to import saved configuration file
        config = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                       srcdir + "/etc/sssd.api.d")
        config.import_config(configfile=of)

        # test set_option 'debug_level' value

        # check internal state before parsing strings which is done in
        # get_domain or get_service
        debug_option = [x for x in config.options('domain/LDAP')
                        if x['name'] == 'debug_level']
        self.assertEqual(len(debug_option), 1)
        self.assertEqual(debug_option[0]['value'], '3')

        debug_option = [x for x in config.options('domain/PROXY')
                        if x['name'] == 'debug_level']
        self.assertEqual(len(debug_option), 1)
        self.assertEqual(debug_option[0]['value'], '0x1f10')

        debug_option = [x for x in config.options('sudo')
                        if x['name'] == 'debug_level']
        self.assertEqual(len(debug_option), 1)
        self.assertEqual(debug_option[0]['value'], '0x2210')

        debug_option = [x for x in config.options('pam')
                        if x['name'] == 'debug_level']
        self.assertEqual(len(debug_option), 1)
        self.assertEqual(debug_option[0]['value'], '9')

        # Remove the output file
        os.unlink(of)

    def testSpaces(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")
        ldap_domain = sssdconfig.get_domain('LDAP')
        self.assertEqual(ldap_domain.get_option('auth_provider'), 'ldap')
        self.assertEqual(ldap_domain.get_option('id_provider'), 'ldap')


class SSSDConfigTestInvalid(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testBadBool(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-invalid-badbool.conf")
        self.assertRaises(TypeError,
                          sssdconfig.get_domain, 'IPA')


class SSSDConfigTestSSSDService(unittest.TestCase):
    def setUp(self):
        self.schema = SSSDConfig.SSSDConfigSchema(srcdir + "/etc/sssd.api.conf",
                                                  srcdir + "/etc/sssd.api.d")

    def tearDown(self):
        pass

    def testInit(self):
        # Positive test
        SSSDConfig.SSSDService('sssd', self.schema)

        # Type Error test
        # Name is not a string
        self.assertRaises(TypeError, SSSDConfig.SSSDService, 3, self.schema)

        # TypeError test
        # schema is not an SSSDSchema
        self.assertRaises(TypeError, SSSDConfig.SSSDService, '3', self)

        # ServiceNotRecognizedError test
        self.assertRaises(SSSDConfig.ServiceNotRecognizedError,
                          SSSDConfig.SSSDService, 'ssd', self.schema)

    def testListOptions(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        options = service.list_options()
        control_list = [
            'services',
            'domains',
            'timeout',
            're_expression',
            'full_name_format',
            'krb5_rcache_dir',
            'user',
            'default_domain_suffix',
            'debug',
            'debug_level',
            'debug_timestamps',
            'debug_microseconds',
            'debug_backtrace_enabled',
            'command',
            'fd_limit',
            'client_idle_timeout',
            'responder_idle_timeout',
            'cache_first',
            'description',
            'certificate_verification',
            'override_space',
            'disable_netlink',
            'enable_files_domain',
            'domain_resolution_order',
            'try_inotify',
            'monitor_resolv_conf',
            'implicit_pac_responder',
            'core_dumpable',
            'passkey_verification',
        ]

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        self.assertTrue(type(options['services']) == tuple,
                        "Option values should be a tuple")

        self.assertTrue(options['services'][0] == list,
                        "services should require an list. "
                        "list_options is requiring a %s" %
                        options['services'][0])

        self.assertTrue(options['services'][1] == str,
                        "services should require a subtype of str. "
                        "list_options is requiring a %s" %
                        options['services'][1])

    def testListMandatoryOptions(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        options = service.list_mandatory_options()
        control_list = [
            'services',
            'domains']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        self.assertTrue(type(options['services']) == tuple,
                        "Option values should be a tuple")

        self.assertTrue(options['services'][0] == list,
                        "services should require an list. "
                        "list_options is requiring a %s" %
                        options['services'][0])

        self.assertTrue(options['services'][1] == str,
                        "services should require a subtype of str. "
                        "list_options is requiring a %s" %
                        options['services'][1])

    def testSetOption(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        # Positive test - Exactly right
        service.set_option('debug_level', 2)
        self.assertEqual(service.get_option('debug_level'), 2)

        # Positive test - Allow converting "safe" values
        service.set_option('debug_level', '2')
        self.assertEqual(service.get_option('debug_level'), 2)

        # Positive test - Remove option if value is None
        service.set_option('debug_level', None)
        self.assertTrue('debug_level' not in service.options.keys())

        # Negative test - Nonexistent Option
        self.assertRaises(SSSDConfig.NoOptionError, service.set_option, 'nosuchoption', 1)

        # Negative test - Incorrect type
        self.assertRaises(TypeError, service.set_option, 'debug_level', 'two')

    def testGetOption(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        # Positive test - List of values
        self.assertEqual(service.get_option('services'), ['nss', 'pam'])

        # Negative Test - Bad Option
        self.assertRaises(SSSDConfig.NoOptionError, service.get_option, 'nosuchoption')

    def testGetAllOptions(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        # Positive test
        options = service.get_all_options()
        control_list = ['services']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

    def testRemoveOption(self):
        service = SSSDConfig.SSSDService('sssd', self.schema)

        # Positive test - Remove an option that exists
        self.assertEqual(service.get_option('services'), ['nss', 'pam'])
        service.remove_option('services')
        self.assertRaises(SSSDConfig.NoOptionError, service.get_option, 'debug_level')

        # Positive test - Remove an option that doesn't exist
        self.assertRaises(SSSDConfig.NoOptionError, service.get_option, 'nosuchentry')
        service.remove_option('nosuchentry')


class SSSDConfigTestSSSDDomain(unittest.TestCase):
    def setUp(self):
        self.schema = SSSDConfig.SSSDConfigSchema(srcdir + "/etc/sssd.api.conf",
                                                  srcdir + "/etc/sssd.api.d")

    def tearDown(self):
        pass

    def testInit(self):
        # Positive Test
        SSSDConfig.SSSDDomain('mydomain', self.schema)

        # Negative Test - Name not a string
        self.assertRaises(TypeError, SSSDConfig.SSSDDomain, 2, self.schema)

        # Negative Test - Schema is not an SSSDSchema
        self.assertRaises(TypeError, SSSDConfig.SSSDDomain, 'mydomain', self)

    def testGetName(self):
        # Positive Test
        domain = SSSDConfig.SSSDDomain('mydomain', self.schema)

        self.assertEqual(domain.get_name(), 'mydomain')

    def testSetActive(self):
        # Positive Test
        domain = SSSDConfig.SSSDDomain('mydomain', self.schema)

        # Should default to inactive
        self.assertFalse(domain.active)
        domain.set_active(True)
        self.assertTrue(domain.active)
        domain.set_active(False)
        self.assertFalse(domain.active)

    def testListOptions(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # First test default options
        options = domain.list_options()
        control_list = [
            'description',
            'enabled',
            'debug',
            'debug_level',
            'debug_timestamps',
            'domain_type',
            'min_id',
            'max_id',
            'timeout',
            'offline_timeout',
            'offline_timeout_max',
            'offline_timeout_random_offset',
            'command',
            'enumerate',
            'cache_credentials',
            'cache_credentials_minimal_first_factor_length',
            'use_fully_qualified_names',
            'ignore_group_members',
            'filter_users',
            'filter_groups',
            'entry_cache_timeout',
            'entry_cache_user_timeout',
            'entry_cache_group_timeout',
            'entry_cache_netgroup_timeout',
            'entry_cache_service_timeout',
            'entry_cache_autofs_timeout',
            'entry_cache_sudo_timeout',
            'entry_cache_ssh_host_timeout',
            'entry_cache_resolver_timeout',
            'refresh_expired_interval',
            'lookup_family_order',
            'account_cache_expiration',
            'dns_resolver_server_timeout',
            'dns_resolver_op_timeout',
            'dns_resolver_timeout',
            'dns_discovery_domain',
            'failover_primary_timeout',
            'dyndns_update',
            'dyndns_update_per_family',
            'dyndns_ttl',
            'dyndns_iface',
            'dyndns_refresh_interval',
            'dyndns_refresh_interval_offset',
            'dyndns_update_ptr',
            'dyndns_force_tcp',
            'dyndns_auth',
            'dyndns_server',
            'dyndns_dot_cacert',
            'dyndns_dot_cert',
            'dyndns_dot_key',
            'subdomain_enumerate',
            'override_gid',
            'case_sensitive',
            'override_homedir',
            'fallback_homedir',
            'homedir_substring',
            'override_shell',
            'default_shell',
            'pwd_expiration_warning',
            'id_provider',
            'auth_provider',
            'access_provider',
            'chpass_provider',
            'sudo_provider',
            'autofs_provider',
            'hostid_provider',
            'subdomains_provider',
            'selinux_provider',
            'session_provider',
            'resolver_provider',
            'realmd_tags',
            'subdomain_refresh_interval',
            'subdomain_refresh_interval_offset',
            'subdomain_inherit',
            'subdomain_homedir',
            'full_name_format',
            're_expression',
            'cached_auth_timeout',
            'auto_private_groups',
            'pam_gssapi_services',
            'pam_gssapi_check_upn',
            'pam_gssapi_indicators_map',
            'refresh_expired_interval',
            'refresh_expired_interval_offset',
            'local_auth_policy']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        self.assertTrue(type(options['max_id']) == tuple,
                        "Option values should be a tuple")

        self.assertTrue(options['max_id'][0] == int,
                        "max_id should require an int. "
                        "list_options is requiring a %s" %
                        options['max_id'][0])

        self.assertTrue(options['max_id'][1] is None,
                        "max_id should not require a subtype. "
                        "list_options is requiring a %s" %
                        options['max_id'][1])

        # Add a provider that has global options and verify that
        # The new options appear.
        domain.add_provider('krb5', 'auth')

        backup_list = control_list[:]
        control_list.extend(
            ['krb5_server',
             'krb5_backup_server',
             'krb5_realm',
             'krb5_kpasswd',
             'krb5_backup_kpasswd',
             'krb5_ccachedir',
             'krb5_ccname_template',
             'krb5_keytab',
             'krb5_validate',
             'krb5_store_password_if_offline',
             'krb5_auth_timeout',
             'krb5_renewable_lifetime',
             'krb5_lifetime',
             'krb5_renew_interval',
             'krb5_use_fast',
             'krb5_fast_principal',
             'krb5_fast_use_anonymous_pkinit',
             'krb5_canonicalize',
             'krb5_use_enterprise_principal',
             'krb5_use_subdomain_realm',
             'krb5_use_kdcinfo',
             'krb5_map_user'])

        options = domain.list_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        control_list.extend(['krb5_kdcip'])

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Remove the auth domain and verify that the options
        # revert to the backup_list
        domain.remove_provider('auth')
        options = domain.list_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in backup_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in backup_list,
                            'Option [%s] unexpectedly found' %
                            option)

    def testListMandatoryOptions(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # First test default options
        options = domain.list_mandatory_options()
        control_list = ['id_provider']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Add a provider that has global options and verify that
        # The new options appear.
        domain.add_provider('krb5', 'auth')

        backup_list = control_list[:]
        control_list.extend(['krb5_realm'])

        options = domain.list_mandatory_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Remove the auth domain and verify that the options
        # revert to the backup_list
        domain.remove_provider('auth')
        options = domain.list_mandatory_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in backup_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in backup_list,
                            'Option [%s] unexpectedly found' %
                            option)

    def testListProviders(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        control_provider_dict = {
            'ipa': ['id', 'auth', 'access', 'chpass', 'sudo', 'autofs',
                    'session', 'hostid', 'subdomains'],
            'ad': ['id', 'auth', 'access', 'chpass', 'sudo', 'autofs',
                   'subdomains', 'resolver'],
            'ldap': ['id', 'auth', 'access', 'chpass', 'sudo', 'autofs',
                     'resolver'],
            'krb5': ['auth', 'access', 'chpass'],
            'proxy': ['id', 'auth', 'chpass'],
            'simple': ['access'],
            'permit': ['access'],
            'deny': ['access']}

        providers = domain.list_providers()

        # Ensure that all of the expected defaults are there
        for provider in control_provider_dict.keys():
            for ptype in control_provider_dict[provider]:
                self.assertTrue(provider in providers)
                self.assertTrue(ptype in providers[provider])

        for provider in providers.keys():
            for ptype in providers[provider]:
                self.assertTrue(provider in control_provider_dict)
                self.assertTrue(ptype in control_provider_dict[provider])

    def testListProviderOptions(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Test looking up a specific provider type
        options = domain.list_provider_options('krb5', 'auth')
        control_list = [
            'krb5_server',
            'krb5_backup_server',
            'krb5_kdcip',
            'krb5_realm',
            'krb5_kpasswd',
            'krb5_backup_kpasswd',
            'krb5_ccachedir',
            'krb5_ccname_template',
            'krb5_keytab',
            'krb5_validate',
            'krb5_store_password_if_offline',
            'krb5_auth_timeout',
            'krb5_renewable_lifetime',
            'krb5_lifetime',
            'krb5_renew_interval',
            'krb5_use_fast',
            'krb5_fast_principal',
            'krb5_fast_use_anonymous_pkinit',
            'krb5_canonicalize',
            'krb5_use_enterprise_principal',
            'krb5_use_subdomain_realm',
            'krb5_use_kdcinfo',
            'krb5_map_user']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Test looking up all provider values
        options = domain.list_provider_options('krb5')
        control_list.extend(['krb5_kpasswd'])

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

    def testAddProvider(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Positive Test
        domain.add_provider('proxy', 'id')

        # Negative Test - No such backend type
        self.assertRaises(SSSDConfig.NoSuchProviderError,
                          domain.add_provider, 'nosuchbackend', 'auth')

        # Negative Test - No such backend subtype
        self.assertRaises(SSSDConfig.NoSuchProviderSubtypeError,
                          domain.add_provider, 'ldap', 'nosuchsubtype')

        # Negative Test - Try to add a second provider of the same type
        self.assertRaises(SSSDConfig.ProviderSubtypeInUse,
                          domain.add_provider, 'ldap', 'id')

    def testRemoveProvider(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # First test default options
        options = domain.list_options()
        control_list = [
            'description',
            'enabled',
            'debug',
            'debug_level',
            'debug_timestamps',
            'domain_type',
            'min_id',
            'max_id',
            'timeout',
            'offline_timeout',
            'offline_timeout_max',
            'offline_timeout_random_offset',
            'command',
            'enumerate',
            'cache_credentials',
            'cache_credentials_minimal_first_factor_length',
            'use_fully_qualified_names',
            'ignore_group_members',
            'filter_users',
            'filter_groups',
            'entry_cache_timeout',
            'entry_cache_user_timeout',
            'entry_cache_group_timeout',
            'entry_cache_netgroup_timeout',
            'entry_cache_service_timeout',
            'entry_cache_autofs_timeout',
            'entry_cache_sudo_timeout',
            'entry_cache_ssh_host_timeout',
            'entry_cache_resolver_timeout',
            'refresh_expired_interval',
            'account_cache_expiration',
            'lookup_family_order',
            'dns_resolver_server_timeout',
            'dns_resolver_op_timeout',
            'dns_resolver_timeout',
            'dns_discovery_domain',
            'failover_primary_timeout',
            'dyndns_update',
            'dyndns_update_per_family',
            'dyndns_ttl',
            'dyndns_iface',
            'dyndns_refresh_interval',
            'dyndns_update_ptr',
            'dyndns_force_tcp',
            'dyndns_auth',
            'dyndns_server',
            'dyndns_dot_cacert',
            'dyndns_dot_cert',
            'dyndns_dot_key',
            'subdomain_enumerate',
            'override_gid',
            'case_sensitive',
            'override_homedir',
            'fallback_homedir',
            'homedir_substring',
            'override_shell',
            'default_shell',
            'pwd_expiration_warning',
            'id_provider',
            'auth_provider',
            'access_provider',
            'chpass_provider',
            'sudo_provider',
            'autofs_provider',
            'hostid_provider',
            'subdomains_provider',
            'selinux_provider',
            'session_provider',
            'resolver_provider',
            'realmd_tags',
            'subdomain_refresh_interval',
            'subdomain_refresh_interval_offset',
            'subdomain_inherit',
            'subdomain_homedir',
            'full_name_format',
            're_expression',
            'cached_auth_timeout',
            'auto_private_groups',
            'pam_gssapi_services',
            'pam_gssapi_check_upn',
            'pam_gssapi_indicators_map',
            'refresh_expired_interval',
            'refresh_expired_interval_offset',
            'dyndns_refresh_interval',
            'dyndns_refresh_interval_offset',
            'local_auth_policy']

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        self.assertTrue(type(options['max_id']) == tuple,
                        "Option values should be a tuple")

        self.assertTrue(options['max_id'][0] == int,
                        "max_id should require an int. "
                        "list_options is requiring a %s" %
                        options['max_id'][0])

        self.assertTrue(options['max_id'][1] is None,
                        "max_id should not require a subtype. "
                        "list_options is requiring a %s" %
                        options['max_id'][1])

        # Add a provider that has global options and verify that
        # The new options appear.
        domain.add_provider('krb5', 'auth')

        backup_list = control_list[:]
        control_list.extend(
            ['krb5_server',
             'krb5_backup_server',
             'krb5_kdcip',
             'krb5_realm',
             'krb5_kpasswd',
             'krb5_backup_kpasswd',
             'krb5_ccachedir',
             'krb5_ccname_template',
             'krb5_keytab',
             'krb5_validate',
             'krb5_store_password_if_offline',
             'krb5_auth_timeout',
             'krb5_renewable_lifetime',
             'krb5_lifetime',
             'krb5_renew_interval',
             'krb5_use_fast',
             'krb5_fast_principal',
             'krb5_fast_use_anonymous_pkinit',
             'krb5_canonicalize',
             'krb5_use_enterprise_principal',
             'krb5_use_subdomain_realm',
             'krb5_use_kdcinfo',
             'krb5_map_user'])

        options = domain.list_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in control_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in control_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Add an LDAP one
        # LDAP ID providers can also use the krb5_realm
        domain.add_provider('ldap', 'id')

        # Set the krb5_realm option and the ldap_uri option
        domain.set_option('krb5_realm', 'EXAMPLE.COM')
        domain.set_option('ldap_uri', 'ldap://ldap.example.com')

        self.assertEqual(domain.get_option('krb5_realm'),
                         'EXAMPLE.COM')
        self.assertEqual(domain.get_option('ldap_uri'),
                         'ldap://ldap.example.com')

        # Remove the LDAP provider and verify that krb5_realm remains
        domain.remove_provider('id')
        self.assertEqual(domain.get_option('krb5_realm'),
                         'EXAMPLE.COM')
        self.assertFalse('ldap_uri' in domain.options)

        # Remove the auth domain and verify that the options
        # revert to the backup_list
        domain.remove_provider('auth')
        self.assertFalse('auth_provider' in domain.options)
        options = domain.list_options()

        self.assertTrue(type(options) == dict,
                        "Options should be a dictionary")

        # Ensure that all of the expected defaults are there
        for option in backup_list:
            self.assertTrue(option in options.keys(),
                            "Option [%s] missing" %
                            option)

        # Ensure that there aren't any unexpected options listed
        for option in options.keys():
            self.assertTrue(option in backup_list,
                            'Option [%s] unexpectedly found' %
                            option)

        # Ensure that the krb5_realm option is now gone
        self.assertFalse('krb5_realm' in domain.options)

        # Test removing nonexistent provider - Real
        domain.remove_provider('id')
        self.assertFalse('id_provider' in domain.options)

        # Test removing nonexistent provider - Bad backend type
        # Should pass without complaint
        domain.remove_provider('id')
        self.assertFalse('id_provider' in domain.options)

        # Test removing nonexistent provider - Bad provider type
        # Should pass without complaint
        domain.remove_provider('nosuchprovider')
        self.assertFalse('nosuchprovider_provider' in domain.options)

    def testGetOption(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Negative Test - Try to get valid option that is not set
        self.assertRaises(SSSDConfig.NoOptionError, domain.get_option, 'max_id')

        # Positive Test - Set the above option and get it
        domain.set_option('max_id', 10000)
        self.assertEqual(domain.get_option('max_id'), 10000)

        # Negative Test - Try yo get invalid option
        self.assertRaises(SSSDConfig.NoOptionError, domain.get_option, 'nosuchoption')

    def testSetOption(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Positive Test
        domain.set_option('max_id', 10000)
        self.assertEqual(domain.get_option('max_id'), 10000)

        # Positive Test - Remove option if value is None
        domain.set_option('max_id', None)
        self.assertTrue('max_id' not in domain.get_all_options().keys())

        # Negative Test - invalid option
        self.assertRaises(SSSDConfig.NoOptionError, domain.set_option, 'nosuchoption', 1)

        # Negative Test - incorrect type
        self.assertRaises(TypeError, domain.set_option, 'max_id', 'a string')

        # Positive Test - Coax options to appropriate type
        domain.set_option('max_id', '10000')
        self.assertEqual(domain.get_option('max_id'), 10000)

        domain.set_option('max_id', 30.2)
        self.assertEqual(domain.get_option('max_id'), 30)

    def testRemoveOption(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Positive test - Remove unset but valid option
        self.assertFalse('max_id' in domain.get_all_options().keys())
        domain.remove_option('max_id')
        self.assertFalse('max_id' in domain.get_all_options().keys())

        # Positive test - Remove unset and unknown option
        self.assertFalse('nosuchoption' in domain.get_all_options().keys())
        domain.remove_option('nosuchoption')
        self.assertFalse('nosuchoption' in domain.get_all_options().keys())

    def testSetName(self):
        domain = SSSDConfig.SSSDDomain('sssd', self.schema)

        # Positive test - Change the name once
        domain.set_name('sssd2')
        self.assertEqual(domain.get_name(), 'sssd2')
        self.assertEqual(domain.oldname, 'sssd')

        # Positive test - Change the name a second time
        domain.set_name('sssd3')
        self.assertEqual(domain.get_name(), 'sssd3')
        self.assertEqual(domain.oldname, 'sssd')

        # Negative test - try setting the name to a non-string
        self.assertRaises(TypeError,
                          domain.set_name, 4)


class SSSDConfigTestSSSDConfig(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = create_temp_dir()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def testInit(self):
        # Positive test
        SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                              srcdir + "/etc/sssd.api.d")

        # Negative Test - No Such File
        self.assertRaises(IOError,
                          SSSDConfig.SSSDConfig, "nosuchfile.api.conf", srcdir + "/etc/sssd.api.d")

        # Negative Test - Schema is not parsable
        self.assertRaises(SSSDConfig.ParsingError,
                          SSSDConfig.SSSDConfig, srcdir + "/testconfigs/noparse.api.conf", srcdir + "/etc/sssd.api.d")

    def testImportConfig(self):
        # Positive Test
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Verify that all sections were imported
        control_list = [
            'sssd',
            'nss',
            'pam',
            'sudo',
            'domain/PROXY',
            'domain/IPA',
            'domain/LDAP',
            'domain/INVALIDPROVIDER',
            'domain/INVALIDOPTION',
        ]

        for section in control_list:
            self.assertTrue(sssdconfig.has_section(section),
                            "Section [%s] missing" %
                            section)
        for section in sssdconfig.sections():
            self.assertTrue(section['name'] in control_list)

        # Verify that all options were imported for a section
        control_list = [
            'services',
            'domains',
            'debug_timestamps']

        for option in control_list:
            self.assertTrue(sssdconfig.has_option('sssd', option),
                            "Option [%s] missing from [sssd]" %
                            option)
        for option in sssdconfig.options('sssd'):
            if option['type'] in ('empty', 'comment'):
                continue
            self.assertTrue(option['name'] in control_list,
                            "Option [%s] unexpectedly found" %
                            option)

        # TODO: Check the types and values of the settings

        # Negative Test - Missing config file
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        self.assertRaises(IOError, sssdconfig.import_config, "nosuchfile.conf")

        # Negative Test - Invalid config file
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        self.assertRaises(SSSDConfig.ParsingError, sssdconfig.import_config, srcdir + "/testconfigs/sssd-invalid.conf")

        # Negative Test - Already initialized
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")
        self.assertRaises(SSSDConfig.AlreadyInitializedError,
                          sssdconfig.import_config, srcdir + "/testconfigs/sssd-valid.conf")

    def testImportConfigNoVersion(self):
        # Positive Test
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(
            srcdir + "/testconfigs/sssd-noversion.conf"
        )

        # Validate services
        services = sssdconfig.list_services()
        self.assertTrue('sssd' in services)
        self.assertTrue('nss' in services)
        self.assertTrue('pam' in services)
        self.assertTrue('dp' in services)

        # Verify service attributes
        sssd_service = sssdconfig.get_service('sssd')
        service_opts = sssd_service.list_options()

        self.assertTrue('services' in service_opts.keys())
        service_list = sssd_service.get_option('services')
        self.assertTrue('nss' in service_list)
        self.assertTrue('pam' in service_list)

        # Validate domain list
        domains = sssdconfig.list_domains()
        self.assertTrue('LDAP' in domains)
        self.assertTrue('PROXY' in domains)
        self.assertTrue('IPA' in domains)

        # Verify domain attributes
        ipa_domain = sssdconfig.get_domain('IPA')
        domain_opts = ipa_domain.list_options()
        self.assertTrue('debug_level' in domain_opts.keys())
        self.assertTrue('id_provider' in domain_opts.keys())
        self.assertTrue('auth_provider' in domain_opts.keys())

        # Verify domain attributes
        proxy_domain = sssdconfig.get_domain('PROXY')
        domain_opts = proxy_domain.list_options()
        self.assertTrue('debug_level' in domain_opts.keys())
        self.assertTrue('id_provider' in domain_opts.keys())
        self.assertTrue('auth_provider' in domain_opts.keys())

        # Verify domain attributes
        ldap_domain = sssdconfig.get_domain('LDAP')
        domain_opts = ldap_domain.list_options()
        self.assertTrue('debug_level' in domain_opts.keys())
        self.assertTrue('id_provider' in domain_opts.keys())
        self.assertTrue('auth_provider' in domain_opts.keys())

        domain_control_list = [
            'cache_credentials',
            'id_provider',
            'auth_provider',
            'access_provider',
            'autofs_provider',
            'chpass_provider',
            'sudo_provider',
            'subdomains_provider',
            'resolver_provider',
            'default_shell',
            'fallback_homedir',
            'cache_credentials',
            'use_fully_qualified_names',
        ]

        ad_domain = sssdconfig.get_domain("ad.example.com")

        for option in ad_domain.get_all_options():
            self.assertTrue(option in domain_control_list)

        negative_domain_control_list = [
            'ad_server',
            'ldap_id_mapping',
            'ldap_sasl_authid',
            'selinux_provider',
            'hostid_provider',
            'session_provider',
        ]

        for option in ad_domain.get_all_options():
            self.assertFalse(option in negative_domain_control_list)

    def testNewConfig(self):
        # Positive Test
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.new_config()

        # Check that the defaults were set
        control_list = [
            'sssd',
            'nss',
            'pam',
            'sudo',
            'autofs',
            'ssh',
            'pac',
            'ifp',
            'session_recording']
        for section in control_list:
            self.assertTrue(sssdconfig.has_section(section),
                            "Section [%s] missing" %
                            section)
        for section in sssdconfig.sections():
            self.assertTrue(section['name'] in control_list)

        control_list = ['services']
        for option in control_list:
            self.assertTrue(sssdconfig.has_option('sssd', option),
                            "Option [%s] missing from [sssd]" %
                            option)
        for option in sssdconfig.options('sssd'):
            if option['type'] in ('empty', 'comment'):
                continue
            self.assertTrue(option['name'] in control_list,
                            "Option [%s] unexpectedly found" %
                            option)

        # Negative Test - Already Initialized
        self.assertRaises(SSSDConfig.AlreadyInitializedError, sssdconfig.new_config)

    def testWrite(self):
        # TODO Write tests to compare output files
        pass

    def testListActiveServices(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_active_services)

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        control_list = [
            'nss',
            'pam']
        active_services = sssdconfig.list_active_services()
        self.assertTrue(isinstance(active_services, list))

        for service in control_list:
            self.assertTrue(service in active_services,
                            "Service [%s] missing" %
                            service)
        for service in active_services:
            self.assertTrue(service in control_list,
                            "Service [%s] unexpectedly found" %
                            service)

    def testListInactiveServices(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_inactive_services)

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        control_list = [
            'sssd',
            'sudo']
        inactive_services = sssdconfig.list_inactive_services()

        for service in control_list:
            self.assertTrue(service in inactive_services,
                            "Service [%s] missing" %
                            service)
        for service in inactive_services:
            self.assertTrue(service in control_list,
                            "Service [%s] unexpectedly found" %
                            service)

    def testListServices(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - sssdconfig not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_services)

        sssdconfig.new_config()

        control_list = [
            'sssd',
            'pam',
            'nss',
            'sudo',
            'autofs',
            'ssh',
            'pac',
            'ifp',
            'session_recording']
        service_list = sssdconfig.list_services()
        for service in control_list:
            self.assertTrue(service in service_list,
                            "Service [%s] missing" %
                            service)
        for service in service_list:
            self.assertTrue(service in control_list,
                            "Service [%s] unexpectedly found" %
                            service)

    def testGetService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.get_service, 'sssd')

        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        service = sssdconfig.get_service('sssd')
        self.assertTrue(isinstance(service, SSSDConfig.SSSDService))

        # Verify the contents of this service
        self.assertEqual(type(service.get_option('debug_timestamps')), bool)
        self.assertFalse(service.get_option('debug_timestamps'))

        # Negative Test - No such service
        self.assertRaises(SSSDConfig.NoServiceError, sssdconfig.get_service, 'nosuchservice')

        # Positive test - Service with invalid option loads
        # but ignores the invalid option
        service = sssdconfig.get_service('pam')
        self.assertFalse('nosuchoption' in service.options)

    def testNewService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.new_service, 'sssd')

        sssdconfig.new_config()

        # Positive Test
        # First need to remove the existing service
        sssdconfig.delete_service('sssd')
        service = sssdconfig.new_service('sssd')
        self.assertTrue(service.get_name() in sssdconfig.list_services())

        # TODO: check that the values of this new service
        # are set to the defaults from the schema

    def testDeleteService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.delete_service, 'sssd')

        sssdconfig.new_config()

        # Positive Test
        sssdconfig.delete_service('sssd')

    def testSaveService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        new_service = SSSDConfig.SSSDService('sssd', sssdconfig.schema)

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.save_service, new_service)

        # Positive Test
        sssdconfig.new_config()
        sssdconfig.save_service(new_service)

        # TODO: check that all entries were saved correctly (change a few)

        # Negative Test - Type Error
        self.assertRaises(TypeError, sssdconfig.save_service, self)

    def testActivateService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        service_name = 'sudo'

        # Negative test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError,
                          sssdconfig.activate_service, service_name)

        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Positive test - Activate an inactive service
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertFalse(service_name in sssdconfig.list_active_services())
        self.assertTrue(service_name in sssdconfig.list_inactive_services())

        sssdconfig.activate_service(service_name)
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertTrue(service_name in sssdconfig.list_active_services())
        self.assertFalse(service_name in sssdconfig.list_inactive_services())

        # Positive test - Activate an active service
        # This should succeed
        sssdconfig.activate_service(service_name)
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertTrue(service_name in sssdconfig.list_active_services())
        self.assertFalse(service_name in sssdconfig.list_inactive_services())

        # Negative test - Invalid service name
        self.assertRaises(SSSDConfig.NoServiceError,
                          sssdconfig.activate_service, 'nosuchservice')

        # Negative test - Invalid service name type
        self.assertRaises(SSSDConfig.NoServiceError,
                          sssdconfig.activate_service, self)

    def testDeactivateService(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        service_name = 'pam'

        # Negative test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError,
                          sssdconfig.activate_service, service_name)

        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Positive test -Deactivate an active service
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertTrue(service_name in sssdconfig.list_active_services())
        self.assertFalse(service_name in sssdconfig.list_inactive_services())

        sssdconfig.deactivate_service(service_name)
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertFalse(service_name in sssdconfig.list_active_services())
        self.assertTrue(service_name in sssdconfig.list_inactive_services())

        # Positive test - Deactivate an inactive service
        # This should succeed
        sssdconfig.deactivate_service(service_name)
        self.assertTrue(service_name in sssdconfig.list_services())
        self.assertFalse(service_name in sssdconfig.list_active_services())
        self.assertTrue(service_name in sssdconfig.list_inactive_services())

        # Negative test - Invalid service name
        self.assertRaises(SSSDConfig.NoServiceError,
                          sssdconfig.activate_service, 'nosuchservice')

        # Negative test - Invalid service name type
        self.assertRaises(SSSDConfig.NoServiceError,
                          sssdconfig.activate_service, self)

    def testListActiveDomains(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_active_domains)

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        control_list = [
            'IPA']
        active_domains = sssdconfig.list_active_domains()
        self.assertTrue(isinstance(active_domains, list))

        for domain in control_list:
            self.assertTrue(domain in active_domains,
                            "Domain [%s] missing" %
                            domain)
        for domain in active_domains:
            self.assertTrue(domain in control_list,
                            "Domain [%s] unexpectedly found" %
                            domain)

    def testListInactiveDomains(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_inactive_domains)

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        control_list = [
            'PROXY',
            'LDAP',
            'INVALIDPROVIDER',
            'INVALIDOPTION',
        ]
        inactive_domains = sssdconfig.list_inactive_domains()

        for domain in control_list:
            self.assertTrue(domain in inactive_domains,
                            "Domain [%s] missing" %
                            domain)
        for domain in inactive_domains:
            self.assertTrue(domain in control_list,
                            "Domain [%s] unexpectedly found" %
                            domain)

    def testListDomains(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.list_domains)

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        control_list = [
            'IPA',
            'PROXY',
            'LDAP',
            'INVALIDPROVIDER',
            'INVALIDOPTION',
        ]
        domains = sssdconfig.list_domains()

        for domain in control_list:
            self.assertTrue(domain in domains,
                            "Domain [%s] missing" %
                            domain)
        for domain in domains:
            self.assertTrue(domain in control_list,
                            "Domain [%s] unexpectedly found" %
                            domain)

    def testListWithInvalidDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not Initialized
        self.assertRaises(SSSDConfig.NotInitializedError,
                          sssdconfig.list_domains)

        # Positive Test
        sssdconfig.import_config(
            srcdir + '/testconfigs/sssd-nonexisting-services-domains.conf'
        )

        domains = sssdconfig.list_active_domains()
        self.assertTrue("active" in domains and len(domains) == 1,
                        "domain 'active' not found among active domains")

        domains = sssdconfig.list_inactive_domains()
        self.assertTrue("inactive" in domains and len(domains) == 1,
                        "domain 'inactive' not found among inactive domains")

        services = sssdconfig.list_active_services()
        self.assertTrue("nss" in services and len(services) == 1,
                        "service 'nss' not found among active services")

        services = sssdconfig.list_inactive_services()
        self.assertTrue(len(services) == 2,
                        "unexpected count of inactive services")
        for service in ("sssd", "pam"):
            self.assertTrue(service in services,
                            "service '%s' not found among inactive services"
                            % service)

    def testGetDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.get_domain, 'sssd')

        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        domain = sssdconfig.get_domain('IPA')
        self.assertTrue(isinstance(domain, SSSDConfig.SSSDDomain))
        self.assertTrue(domain.active)

        domain = sssdconfig.get_domain('LDAP')
        self.assertTrue(isinstance(domain, SSSDConfig.SSSDDomain))
        self.assertFalse(domain.active)

        # TODO verify the contents of this domain
        self.assertTrue(domain.get_option('ldap_id_use_start_tls'))
        self.assertTrue(domain.get_option('ldap_sudo_include_regexp'))
        self.assertTrue(domain.get_option('ldap_autofs_map_master_name'))

        # Negative Test - No such domain
        self.assertRaises(SSSDConfig.NoDomainError, sssdconfig.get_domain, 'nosuchdomain')

        # Positive Test - Domain with unknown provider
        # Expected result: Domain is imported, but does not contain the
        # unknown provider entry
        domain = sssdconfig.get_domain('INVALIDPROVIDER')
        self.assertFalse('chpass_provider' in domain.options)

        # Positive Test - Domain with unknown option
        # Expected result: Domain is imported, but does not contain the
        # unknown option entry
        domain = sssdconfig.get_domain('INVALIDOPTION')
        self.assertFalse('nosuchoption' in domain.options)

    def testNewDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.new_domain, 'example.com')

        sssdconfig.new_config()

        # Positive Test
        domain = sssdconfig.new_domain('example.com')
        self.assertTrue(isinstance(domain, SSSDConfig.SSSDDomain))
        self.assertTrue(domain.get_name() in sssdconfig.list_domains())
        self.assertTrue(domain.get_name() in sssdconfig.list_inactive_domains())

        # TODO: check that the values of this new domain
        # are set to the defaults from the schema

    def testDeleteDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.delete_domain, 'IPA')

        # Positive Test
        sssdconfig.import_config(srcdir + '/testconfigs/sssd-valid.conf')

        self.assertTrue('IPA' in sssdconfig.list_domains())
        self.assertTrue('IPA' in sssdconfig.list_active_domains())
        self.assertTrue(sssdconfig.has_section('domain/IPA'))
        sssdconfig.delete_domain('IPA')
        self.assertFalse('IPA' in sssdconfig.list_domains())
        self.assertFalse('IPA' in sssdconfig.list_active_domains())
        self.assertFalse(sssdconfig.has_section('domain/IPA'))

    def testSaveDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        # Negative Test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError, sssdconfig.save_domain, 'IPA')

        # Positive Test
        sssdconfig.new_config()
        domain = sssdconfig.new_domain('example.com')
        domain.add_provider('ldap', 'id')
        domain.set_option('ldap_uri', 'ldap://ldap.example.com')
        domain.set_active(True)
        sssdconfig.save_domain(domain)

        self.assertTrue('example.com' in sssdconfig.list_domains())
        self.assertTrue('example.com' in sssdconfig.list_active_domains())
        self.assertEqual(sssdconfig.get('domain/example.com', 'ldap_uri'),
                         'ldap://ldap.example.com')

        # Negative Test - Type Error
        self.assertRaises(TypeError, sssdconfig.save_domain, self)

        # Positive test - Change the domain name and save it
        domain.set_name('example.com2')
        self.assertEqual(domain.name, 'example.com2')
        self.assertEqual(domain.oldname, 'example.com')
        sssdconfig.save_domain(domain)

        self.assertTrue('example.com2' in sssdconfig.list_domains())
        self.assertTrue('example.com2' in sssdconfig.list_active_domains())
        self.assertTrue(sssdconfig.has_section('domain/example.com2'))
        self.assertEqual(sssdconfig.get('domain/example.com2',
                                        'ldap_uri'),
                         'ldap://ldap.example.com')
        self.assertFalse('example.com' in sssdconfig.list_domains())
        self.assertFalse('example.com' in sssdconfig.list_active_domains())
        self.assertFalse('example.com' in sssdconfig.list_inactive_domains())
        self.assertFalse(sssdconfig.has_section('domain/example.com'))
        self.assertEqual(domain.oldname, None)

        # Positive test - Set the domain inactive and save it
        activelist = sssdconfig.list_active_domains()
        inactivelist = sssdconfig.list_inactive_domains()

        domain.set_active(False)
        sssdconfig.save_domain(domain)

        self.assertFalse('example.com2' in sssdconfig.list_active_domains())
        self.assertTrue('example.com2' in sssdconfig.list_inactive_domains())

        self.assertEqual(len(sssdconfig.list_active_domains()),
                         len(activelist) - 1)
        self.assertEqual(len(sssdconfig.list_inactive_domains()),
                         len(inactivelist) + 1)

        # Positive test - Set the domain active and save it
        activelist = sssdconfig.list_active_domains()
        inactivelist = sssdconfig.list_inactive_domains()
        domain.set_active(True)
        sssdconfig.save_domain(domain)

        self.assertTrue('example.com2' in sssdconfig.list_active_domains())
        self.assertFalse('example.com2' in sssdconfig.list_inactive_domains())

        self.assertEqual(len(sssdconfig.list_active_domains()),
                         len(activelist) + 1)
        self.assertEqual(len(sssdconfig.list_inactive_domains()),
                         len(inactivelist) - 1)

        # Positive test - Set the domain inactive and save it
        activelist = sssdconfig.list_active_domains()
        inactivelist = sssdconfig.list_inactive_domains()

        sssdconfig.deactivate_domain(domain.get_name())

        self.assertFalse('example.com2' in sssdconfig.list_active_domains())
        self.assertTrue('example.com2' in sssdconfig.list_inactive_domains())

        self.assertEqual(len(sssdconfig.list_active_domains()),
                         len(activelist) - 1)
        self.assertEqual(len(sssdconfig.list_inactive_domains()),
                         len(inactivelist) + 1)

        # Positive test - Set the domain active and save it
        activelist = sssdconfig.list_active_domains()
        inactivelist = sssdconfig.list_inactive_domains()

        sssdconfig.activate_domain(domain.get_name())

        self.assertTrue('example.com2' in sssdconfig.list_active_domains())
        self.assertFalse('example.com2' in sssdconfig.list_inactive_domains())

        self.assertEqual(len(sssdconfig.list_active_domains()),
                         len(activelist) + 1)
        self.assertEqual(len(sssdconfig.list_inactive_domains()),
                         len(inactivelist) - 1)

        # Positive test - Ensure that saved domains retain values
        domain.set_option('ldap_krb5_init_creds', True)
        domain.set_option('ldap_id_use_start_tls', False)
        domain.set_option('ldap_user_search_base',
                          'cn=accounts, dc=example, dc=com')
        self.assertTrue(domain.get_option('ldap_krb5_init_creds'))
        self.assertFalse(domain.get_option('ldap_id_use_start_tls'))
        self.assertEqual(domain.get_option('ldap_user_search_base'),
                         'cn=accounts, dc=example, dc=com')

        sssdconfig.save_domain(domain)

        of = self.tmp_dir + '/testSaveDomain.out'

        # Ensure the output file doesn't exist
        try:
            os.unlink(of)
        except OSError:
            pass

        # Write out the file
        sssdconfig.write(of)

        # Verify that the output file has the correct permissions
        mode = os.stat(of)[ST_MODE]

        # Output files should not be readable or writable by
        # others, and should not be executable by anyone
        self.assertFalse(S_IMODE(mode) & 0o137)

        # Remove the output file
        os.unlink(of)

        domain2 = sssdconfig.get_domain('example.com2')
        self.assertTrue(domain2.get_option('ldap_krb5_init_creds'))
        self.assertFalse(domain2.get_option('ldap_id_use_start_tls'))

    def testActivateDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        domain_name = 'PROXY'

        # Negative test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError,
                          sssdconfig.activate_domain, domain_name)

        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Positive test - Activate an inactive domain
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertFalse(domain_name in sssdconfig.list_active_domains())
        self.assertTrue(domain_name in sssdconfig.list_inactive_domains())

        sssdconfig.activate_domain('PROXY')
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertTrue(domain_name in sssdconfig.list_active_domains())
        self.assertFalse(domain_name in sssdconfig.list_inactive_domains())

        # Positive test - Activate an active domain
        # This should succeed
        sssdconfig.activate_domain('PROXY')
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertTrue(domain_name in sssdconfig.list_active_domains())
        self.assertFalse(domain_name in sssdconfig.list_inactive_domains())

        # Negative test - Invalid domain name
        self.assertRaises(SSSDConfig.NoDomainError,
                          sssdconfig.activate_domain, 'nosuchdomain')

        # Negative test - Invalid domain name type
        self.assertRaises(SSSDConfig.NoDomainError,
                          sssdconfig.activate_domain, self)

    def testDeactivateDomain(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        domain_name = 'IPA'

        # Negative test - Not initialized
        self.assertRaises(SSSDConfig.NotInitializedError,
                          sssdconfig.activate_domain, domain_name)

        sssdconfig.import_config(srcdir + "/testconfigs/sssd-valid.conf")

        # Positive test -Deactivate an active domain
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertTrue(domain_name in sssdconfig.list_active_domains())
        self.assertFalse(domain_name in sssdconfig.list_inactive_domains())

        sssdconfig.deactivate_domain(domain_name)
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertFalse(domain_name in sssdconfig.list_active_domains())
        self.assertTrue(domain_name in sssdconfig.list_inactive_domains())

        # Positive test - Deactivate an inactive domain
        # This should succeed
        sssdconfig.deactivate_domain(domain_name)
        self.assertTrue(domain_name in sssdconfig.list_domains())
        self.assertFalse(domain_name in sssdconfig.list_active_domains())
        self.assertTrue(domain_name in sssdconfig.list_inactive_domains())

        # Negative test - Invalid domain name
        self.assertRaises(SSSDConfig.NoDomainError,
                          sssdconfig.activate_domain, 'nosuchdomain')

        # Negative test - Invalid domain name type
        self.assertRaises(SSSDConfig.NoDomainError,
                          sssdconfig.activate_domain, self)

    def testParse(self):
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")

        with open(srcdir + "/testconfigs/sssd-test-parse.conf", "r") as f:
            data = sssdconfig.parse(f)

        self.assertEqual(len(data), 4)
        self.assertEqual(data[-1], {'type': "section",
                                    'name': "nss",
                                    'value': [{'type': 'option',
                                               'name': 'debug_level',
                                               'value': '1'},
                                              {'type': 'empty',
                                               'name': 'empty'}]})

        with open(srcdir + "/testconfigs/sssd-valid.conf", "r") as f:
            data = sssdconfig.parse(f)

        self.assertEqual(len(data), 9)
        self.assertEqual(data[-1], {'name': "sudo",
                                    'type': "section",
                                    'value': [{'type': 'option',
                                               'name': 'debug_level',
                                               'value': '0xfC10'}]})

    def testEnabledOption(self):
        """Test the new enabled option."""
        # Positive Test
        sssdconfig = SSSDConfig.SSSDConfig(srcdir + "/etc/sssd.api.conf",
                                           srcdir + "/etc/sssd.api.d")
        sssdconfig.import_config(srcdir + "/testconfigs/sssd-enabled-option.conf")

        # Verify that all sections were imported
        control_list = [
            'nss',
            'sssd',
            'pam',
            'domain/enabled_1',
            'domain/enabled_2',
            'domain/enabled_3',
            'domain/disabled_1',
            'domain/disabled_2',
            'domain/disabled_3',
        ]

        for section in control_list:
            self.assertTrue(sssdconfig.has_section(section),
                            "Section [%s] missing" %
                            section)
        for section in sssdconfig.sections():
            self.assertTrue(section['name'] in control_list)

        # Verify that all options were imported for [sssd] section
        control_list = [
            'services',
            'domains',
            'debug_timestamps']

        for option in control_list:
            self.assertTrue(sssdconfig.has_option('sssd', option),
                            "Option [%s] missing from [sssd]" %
                            option)
        for option in sssdconfig.options('sssd'):
            if option['type'] in ('empty', 'comment'):
                continue
            self.assertTrue(option['name'] in control_list,
                            "Option [%s] unexpectedly found" %
                            option)

        # Verify enabled domains
        control_list = [
            'enabled_1',
            'enabled_2',
            'enabled_3']

        if (sssdconfig.has_option('sssd', 'domains')):
            sssd_domains = striplist(sssdconfig.get('sssd', 'domains').split(','))
            domain_dict = dict.fromkeys(sssd_domains)
            if '' in domain_dict:
                del domain_dict['']
            sssd_domains = list(domain_dict)
        else:
            sssd_domains = []

        for domain in sssdconfig.list_active_domains():
            self.assertTrue(domain in control_list,
                            "Domain [domain/%s] should be disabled" % domain)
        for domain in control_list:
            self.assertTrue(domain in sssdconfig.list_active_domains(),
                            "Domain [domain/%s] should be enabled" % domain)

        # Verify disabled domains
        control_list = [
            'disabled_1',
            'disabled_2',
            'disabled_3']

        for domain in sssdconfig.list_inactive_domains():
            self.assertTrue(domain in control_list,
                            "Domain [domain/%s] should be enabled" % domain)
        for domain in control_list:
            self.assertTrue(domain in sssdconfig.list_inactive_domains(),
                            "Domain [domain/%s] should be disabled" % domain)


if __name__ == "__main__":
    error = 0

    suite = unittest.TestLoader().loadTestsFromTestCase(SSSDConfigTestSSSDService)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x1

    suite = unittest.TestLoader().loadTestsFromTestCase(SSSDConfigTestSSSDDomain)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x2

    suite = unittest.TestLoader().loadTestsFromTestCase(SSSDConfigTestSSSDConfig)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x4

    suite = unittest.TestLoader().loadTestsFromTestCase(SSSDConfigTestValid)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x8

    suite = unittest.TestLoader().loadTestsFromTestCase(SSSDConfigTestInvalid)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x10

    sys.exit(error)



#
#   Authors:
#       Pavel Brezina <pbrezina@redhat.com>
#
#   Copyright (C) 2017 Red Hat
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


class DataType:
    """ Represents an SBus data type and its corresponding equivalents
        in D-Bus and C code.

        SBus supports also custom types that can be parsed complex C types such
        as hash tables or structures. In this case the SBus type may differ
        from D-Bus type.
    """
    available = {}

    def __init__(self, sbus_type, dbus_type, c_type, key_format,
                 require_talloc):
        self.sbus_type = sbus_type
        self.dbus_type = dbus_type
        self.RequireTalloc = require_talloc

        # Printf formatter (without leading %) if the type supports keying
        self.keyFormat = key_format

        # Input and output C types. For example 'int' and 'int*'
        self.CType = c_type
        self.inputCType = c_type
        self.outputCType = c_type + "*"

    def __del__(self):
        del DataType.available[self.sbus_type]

    def __str__(self):
        return "%s == %s == %s" % (self.sbus_type, self.dbus_type, self.c_type)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def Find(sbus_type):
        """ Find DataType object of given @sbus_type.
        """
        if not (sbus_type in DataType.available):
            raise ValueError(('Data type "%s" is not currently '
                              'supported by code generator') % sbus_type)

        return DataType.available[sbus_type]

    @staticmethod
    def Create(sbus_type, c_type, KeyFormat=None, DBusType=None,
               RequireTalloc=False):
        """ Create a new SBus type. Specify DBusType if it differs from
            the SBus type. Specify printf formatter KeyFormat if this type
            can be used as a key.
        """
        dbus_type = DBusType if DBusType is not None else sbus_type

        type = DataType(sbus_type, dbus_type, c_type, KeyFormat, RequireTalloc)
        DataType.available[sbus_type] = type

        return type

    @staticmethod
    def SBusToDBusType(sbus_type):
        """ If possible convert SBus data type into D-Bus type. Otherwise
            return unchanged input.
        """
        if not (sbus_type in DataType.available):
            return sbus_type

        return DataType.available[sbus_type].dbus_type



#!/usr/bin/env python3
from __future__ import print_function
from datetime import datetime
import os
import subprocess
import argparse
import shlex
import errno


class LookupPerf(object):
    def __init__(self, database, name, maxlookup, datadir):
        self.database = database
        self.lookupname = name
        self.lookup_max = maxlookup
        self.perf_data_dir = datadir
        self.mem_cache_data_dir = '%s/memory_cache' % (self.perf_data_dir)
        self.ldb_cache_dir = '%s/ldb_cache' % (self.perf_data_dir)
        if self.database == 'passwd':
            self.identity = 'users'
        elif self.database == 'group':
            self.identity = 'group'
        elif self.database == 'initgroups':
            self.identity = 'initgroups'

    def prepare(self):
        """ Create directory structure """
        dirs = [self.perf_data_dir, self.mem_cache_data_dir,
                self.ldb_cache_dir]
        for directory in dirs:
            try:
                os.makedirs(directory)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

    def find_nss(self, tracefile, groupname, loop_count):
        """ Find if nss socket in trace file """
        f = open(tracefile, 'r')
        lines = f.read()
        nss = lines.find('/var/lib/sss/pipes/nss')
        if nss != -1:
            move_trace_file = "mv -f %s %s" % (tracefile, self.ldb_cache_dir)
            self.execute(shlex.split(move_trace_file))
            cp_mc = "cp -a /var/lib/sss/mc %s" % (self.perf_data_dir)
            self.execute(shlex.split(cp_mc))
            print("%s was looked through nss socket" % groupname)
            print("Total %s stored in memcache is %d" % (self.identity,
                                                         loop_count))
            return True
        else:
            return False

    def run(self):
        """ Run performance loop """
        if self.database != 'initgroups':
            lookup_cmd = 'getent %s' % (self.database)
        else:
            lookup_cmd = 'id'
        for i in range(1, int(self.lookup_max)):
            print("--------------loop %d---------------" % (i))
            ug_name = '%s%d' % (self.lookupname, i)
            getent_cmd = '%s %s' % (lookup_cmd, ug_name)
            print(getent_cmd)
            now = datetime.now()
            timestamp = now.strftime("%H:%M:%S:%f")[:-3]
            _, _, _ = self.execute(shlex.split(getent_cmd))
            print("i = %d, %s, %s" % (i, ug_name, timestamp))
            for j in range(1, min(10, i)):
                ug_name = '%s%d' % (self.lookupname, j)
                strace_file = '%s/%s.trace' % (self.mem_cache_data_dir,
                                               ug_name)
                strace_cmd = 'strace -fxvto %s %s %s' % (strace_file,
                                                         lookup_cmd, ug_name)
                _, _, rc = self.execute(shlex.split(strace_cmd))
                now = datetime.now()
                timestamp = now.strftime("%H:%M:%S:%f")[:-3]
                print("j = %d, %s, %s" % (j, ug_name, timestamp))
                if self.find_nss(strace_file, ug_name, i):
                    return (i)
            k1 = int(i / 5)
            k2 = int((2 * i) / 5)
            k3 = int((3 * i) / 5)
            k4 = int((4 * i) / 5)
            if i > 10:
                my_list = [k1, k2, k3, k4]
                for k in my_list:
                    ug_name = '%s%d' % (self.lookupname, k)
                    strace_file = '%s/%s.trace' % (self.mem_cache_data_dir,
                                                   ug_name)
                    strace_cmd = 'strace -fxvto %s %s %s' % (strace_file,
                                                             lookup_cmd,
                                                             ug_name)
                    _, _, rc = self.execute(shlex.split(strace_cmd))
                    now = datetime.now()
                    timestamp = now.strftime("%H:%M:%S:%f")[:-3]
                    print("k = %d, %s, %s" % (k, ug_name, timestamp))
                    if self.find_nss(strace_file, ug_name, i):
                        return i
                a1 = max(i - 10, 0)
                a2 = i
                for x in range(a1, a2):
                    ug_name = '%s%d' % (self.lookupname, x)
                    strace_file = '%s/%s.trace' % (self.mem_cache_data_dir,
                                                   ug_name)
                    strace_cmd = 'strace -fxvto %s %s %s' % (strace_file,
                                                             lookup_cmd,
                                                             ug_name)
                    _, _, _ = self.execute(shlex.split(strace_cmd))
                    now = datetime.now()
                    timestamp = now.strftime("%H:%M:%S:%f")[:-3]
                    print("x = %d, %s, %s" % (x, ug_name, timestamp))
                    if self.find_nss(strace_file, ug_name, i):
                        return i
        return i

    def execute(self, args, shell=False, stdin=None, capture_output=True,
                raiseonerr=False, env=None, cwd=None,):
        """ Execute command """
        p_in = None
        p_out = None
        p_err = None
        if env is None:
            env = os.environ.copy()
        if capture_output:
            p_out = subprocess.PIPE
            p_err = subprocess.PIPE
        try:
            proc = subprocess.Popen(args, shell=shell, stdin=p_in,
                                    stdout=p_out,
                                    stderr=p_err, close_fds=True,
                                    env=env, cwd=cwd)
            stdout, stderr = proc.communicate(stdin)
        except KeyboardInterrupt:
            proc.wait()
            raise
        if proc.returncode != 0 and raiseonerr:
            raise subprocess.CalledProcessError(proc.returcode, args, stdout)
        else:
            return (stdout, stderr, proc.returncode)


def main():
    parser = argparse.ArgumentParser("Description=Mem Cache Performance")
    parser.add_argument_group('Mandatory Arguments')
    parser.add_argument('database', type=str,
                        help="Specify database(passwd/group/initgroups)",
                        choices=['passwd', 'group', 'initgroups'])
    parser.add_argument('--name', type=str,
                        help="Pass the user or group name to be looked up.\
                        Ex.If usernames are foobar1..N,\
                        specify foobar as username",
                        required=True)
    parser.add_argument('--datadir', type=str,
                        help="Pass Directory path to save performance data",
                        required=True)
    parser.add_argument('--maxlookup', type=int, help="Maximum lookup",
                        required=True)
    args = parser.parse_args()
    print("args =", args)
    e = LookupPerf(args.database, args.name, args.maxlookup, args.datadir)
    e.prepare()
    loop_count = e.run()
    output_file = "%s/%s" % (args.datadir, 'output')
    with open(output_file, 'w+') as f:
        f.write("%d\n" % loop_count)


if __name__ == '__main__':
    main()



""" Automation of proxy provider suite

:requirement: IDM-SSSD-REQ : Proxy Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
"""
from __future__ import print_function
import pytest
import subprocess
import time
from sssd.testlib.common.utils import sssdTools, LdapOperations
from sssd.testlib.common.ssh2_python import check_login_client
from constants import ds_suffix, ds_instance_name


def execute_cmd(multihost, command):
    """ Execute command on client """
    cmd = multihost.client[0].run_command(command)
    return cmd


@pytest.fixture(scope='class')
def ldap_objects_sssd_client(multihost, request):
    """
        Configure sssd.conf
        Create a dedicated user with
        a DN starting e.g. with cn=...
        Create a local user foo12
    """
    ldap_uri = 'ldap://%s' % multihost.master[0].sys_hostname
    ds_rootdn = 'cn=Directory Manager'
    ds_rootpw = 'Secret123'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    user_info = {'objectClass': [b'top', b'organizationalUnit'],
                 'ou': 'Users'.encode('utf-8')}
    users_dn = 'ou=Users,dc=example,dc=test'
    (_, _) = ldap_inst.add_entry(user_info, users_dn)
    user_info = {'cn': 'User_CS1'.encode('utf-8'),
                 'homeDirectory': '/home/User_CS1'.encode('utf-8'),
                 'objectClass': [b'account',
                                 b'posixAccount'],
                 'uidNumber': '111111'.encode('utf-8'),
                 'gidNumber': '111111'.encode('utf-8'),
                 'loginShell': '/bin/bash'.encode('utf-8'),
                 'userPassword': 'Secret123'.encode('utf-8'),
                 'uid': ['User_CS1'.encode('utf-8'),
                         'User_CS1_Alias'.encode('utf-8')]}
    user_dn = f'uid=User_CS1,ou=Users,{ds_suffix}'
    (_, _) = ldap_inst.add_entry(user_info, user_dn)
    grop_info = {'cn': ['User_CS1_grp1'.encode('utf-8'),
                        'User_CS1_grp1_Alias'.encode('utf-8')],
                 'objectClass': b'posixGroup',
                 'gidNumber': '111111'.encode('utf-8'),
                 'memberUid': 'User_CS1'.encode('utf-8')}
    group_dn = f'cn=User_CS1_grp1,ou=Groups,{ds_suffix}'
    (_, _) = ldap_inst.add_entry(grop_info, group_dn)
    execute_cmd(multihost, "> /etc/pam_ldap.conf")
    execute_cmd(multihost, f"echo 'base {ds_suffix}' > "
                           f"/etc/pam_ldap.conf")
    execute_cmd(multihost, "echo 'pam_password md5'"
                           " >> /etc/pam_ldap.conf")
    execute_cmd(multihost, f"echo 'host {multihost.master[0].ip}' "
                           f">> /etc/pam_ldap.conf")
    execute_cmd(multihost, "echo 'tls_cacertfile "
                           "/etc/openldap/certs/cacert.asc' >> "
                           "/etc/pam_ldap.conf")
    """
        Interaction between nss-pam-ldapd and a feature of pam_usertype.
        pam_usertype might send a lookup for a user called
        'pam_usertype_non_existent:' and nss-pam-ldapd does not like the ':' in
        the username.

        To fix this we need to add following for RHEL9 and Fedora
    """
    version = "cat /etc/redhat-release"
    if '9.' or 'Fedora' in execute_cmd(multihost, version).stdout_text:
        execute_cmd(multihost, "echo 'validnames /^[a-z0-9._@$()]([a-z0-9._@$()"
                               " \\~-]*[a-z:0-9._@$()~-])?$/i' >> /etc/nslcd.conf")
    execute_cmd(multihost, 'systemctl restart nslcd')

    def restore_user_grp():
        """ Restore sssd.conf """
        ldap_inst.del_dn(f'uid=User_CS1,ou=Users,{ds_suffix}')
        ldap_inst.del_dn(f'cn=User_CS1_grp1,ou=Groups,{ds_suffix}')
        ldap_inst.del_dn(users_dn)
        execute_cmd(multihost, "rm -vf /etc/pam_ldap.conf")

    request.addfinalizer(restore_user_grp)


def config_sssd(multihost):
    """
    Configure common parameters for sssd.conf
    """
    tools = sssdTools(multihost.client[0])
    sssd_params = {'domains': ds_instance_name}
    tools.sssd_conf('sssd', sssd_params)
    domain_name = tools.get_domain_section_name()
    domain_params = {'debug_level': '9',
                     'id_provider': 'proxy',
                     'proxy_lib_name': 'ldap',
                     'proxy_pam_target': 'sssdproxyldap'}
    tools.sssd_conf('domain/' + domain_name, domain_params)


@pytest.mark.usefixtures('setupds',
                         'default_sssd',
                         'sssdproxyldap',
                         'install_nslcd',
                         'ldap_objects_sssd_client')
@pytest.mark.tier1_3
class TestProxyrfc2307(object):
    """
    This is test case class for proxy provider suite
    """
    def test_lookup_user_group(self, multihost, backupsssdconf):
        """
        :title: case sensitive is true lookup user group
        :id: b7a0e1b6-4332-11ed-9a72-845cf3eff344
        :steps:
          1. While case_sensitive = true, lookup user group
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        config_sssd(multihost)
        tools.clear_sssd_cache()
        for i in ["id User_CS1 | grep User_CS1_grp1",
                  "id User_CS1_Alias | grep User_CS1_grp1",
                  "getent passwd User_CS1 | grep User_CS1",
                  "getent passwd User_CS1_Alias | grep User_CS1",
                  "getent group User_CS1_grp1 | grep User_CS1_grp1",
                  "getent group User_CS1_grp1 | grep User_CS1",
                  "getent group User_CS1_grp1_Alias | grep User_CS1_grp1",
                  "getent group User_CS1_grp1_Alias | grep User_CS1"]:
            execute_cmd(multihost, i)
        for user in ['User_CS1', 'User_CS1_Alias']:
            check_login_client(multihost, user, "Secret123")
        for i in ["getent passwd user_cs1",
                  "getent passwd user_cs1_alias",
                  "getent group user_cs1_grp1",
                  "getent group user_cs1_grp1"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, i)

    def test_enumerate_users_groups(self, multihost, backupsssdconf):
        """
        :title: enumerate users groups
        :id: c8ad3f72-4332-11ed-8272-845cf3eff344
        :steps:
          1. While enumerate=True, lookup user group
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'enumerate': 'true'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent -s sss passwd | grep User_CS1",
                  "getent -s sss group | grep User_CS1_grp1"]:
            execute_cmd(multihost, i)

    def test_simple_deny_groups_user_cs1_grp1(self, multihost, backupsssdconf):
        """
        :title: simple deny groups is User CS1 grp1
        :id: cf747f6e-4332-11ed-89fd-845cf3eff344
        :caseposneg: negative
        :steps:
          1. While case_sensitive=True, simple deny groups is User CS1 grp1
        :expectedresults:
          1. Should not succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_deny_groups': 'User_CS1_grp1,User_CS1_grp1_Alias'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent group User_CS1_grp1",
                  "id User_CS1",
                  "> /var/log/secure"]:
            execute_cmd(multihost, i)
        with pytest.raises(Exception):
            check_login_client(multihost, "User_CS1", "Secret123")
        time.sleep(3)
        execute_cmd(multihost, 'cat /var/log/secure | grep -i "Access denied for user User_CS1"')

    def test_simple_deny_groups_user_cs_grp1(self, multihost, backupsssdconf):
        """
        :title: simple deny groups is user cs1 grp1
        :id: d509cf4c-4332-11ed-a3b2-845cf3eff344
        :steps:
          1. While case_sensitive=True, simple deny groups is user cs1 grp1
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_deny_groups': "user_cs1_grp1"}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent passwd User_CS1", "id User_CS1"]:
            execute_cmd(multihost, i)
        check_login_client(multihost, "User_CS1", "Secret123")

    def test_simple_deny_users_user_CS1(self, multihost, backupsssdconf):
        """
        :title: simple deny users is User CS1
        :id: de48ec6e-4332-11ed-aa31-845cf3eff344
        :steps:
          1. While case_sensitive=True, simple deny users is User CS1
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_deny_users': 'User_CS1,User_CS1_Alias'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent group User_CS1_grp1",
                  "id User_CS1",
                  "> /var/log/secure"]:
            execute_cmd(multihost, i)
        with pytest.raises(Exception):
            check_login_client(multihost, "User_CS1", "Secret123")
        time.sleep(3)
        execute_cmd(multihost, 'cat /var/log/secure | grep -i "Access denied for user User_CS1"')

    def test_simple_deny_users_user_cs1(self, multihost, backupsssdconf):
        """
        :title: simple deny users is user cs1
        :id: e5387c92-4332-11ed-897f-845cf3eff344
        :steps:
          1. While case_sensitive=True, simple deny users is user cs1
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'true',
                         'simple_deny_users': 'user_cs1'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        check_login_client(multihost, "User_CS1", "Secret123")

    def test_bz1007381(self, multihost, backupsssdconf):
        """
        :title: proxy provider: id lookup shows "Memory buffer error" in domain log
        :id: ead0c6d2-4332-11ed-b33a-845cf3eff344
        :caseposneg: negative
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1007381
        :steps:
          1. id lookup shows "Memory buffer error" in domain log
        :expectedresults:
          1. Should not succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        tools.clear_sssd_cache()
        execute_cmd(multihost, "id User_CS1")
        with pytest.raises(subprocess.CalledProcessError):
            execute_cmd(multihost, "grep -i 'Memory buffer' /var/log/sssd/*")

    def test_negative_cache(self, multihost):
        """
        :title: negative cache test
        :id: f0e7d9ac-4332-11ed-a5b1-845cf3eff344
        :steps:
          1. Wait for negative cache to expire - default 15 seconds
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        tools.clear_sssd_cache()
        for command in ["getent -s sss passwd newuser",
                        "getent -s sss group newgroup"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, command)
        ldap_uri = 'ldap://%s' % multihost.master[0].sys_hostname
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_info = {'cn': 'newuser'.encode('utf-8'),
                     'homeDirectory': '/home/newuser'.encode('utf-8'),
                     'objectClass': [b'account',
                                     b'posixAccount'],
                     'uidNumber': '123456'.encode('utf-8'),
                     'gidNumber': '123456'.encode('utf-8'),
                     'userPassword': 'Secret123'.encode('utf-8'),
                     'uid': 'newuser'.encode('utf-8')}
        user_dn = f'uid=newuser,ou=Users,{ds_suffix}'
        (_, _) = ldap_inst.add_entry(user_info, user_dn)
        grop_info = {'cn': 'newgroup'.encode('utf-8'),
                     'objectClass': b'posixGroup',
                     'gidNumber': '123456'.encode('utf-8'),
                     'memberUid': 'newuser'.encode('utf-8')}
        group_dn = f'cn=newgroup,ou=Groups,{ds_suffix}'
        (_, _) = ldap_inst.add_entry(grop_info, group_dn)
        for command in ["getent -s sss passwd newuser",
                        "getent -s sss group newgroup"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, command)
        # Waiting for negative cache to expire - default 15 seconds
        time.sleep(16)
        for command in ["getent -s sss passwd newuser",
                        "getent -s sss group newgroup"]:
            execute_cmd(multihost, command)
        ldap_inst.del_dn(f'uid=newuser,ou=Users,{ds_suffix}')
        ldap_inst.del_dn(f'cn=newgroup,ou=Groups,{ds_suffix}')

    def test_nested_group(self, multihost, backupsssdconf):
        """
        :title: nested group
        :id: f76c3fe8-4332-11ed-bce7-845cf3eff344
        :steps:
          1. Look for nested group/user
        :expectedresults:
          1. Should succeed
        """
        ldap_uri = 'ldap://%s' % multihost.master[0].sys_hostname
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        for (cn, uid) in [('user1', '111112'), ('user2', '111113')]:
            user_info = {'cn': cn.encode('utf-8'),
                         'homeDirectory': f'/export/{cn}'.encode('utf-8'),
                         'objectClass': [b'account',
                                         b'posixAccount'],
                         'uidNumber': uid.encode('utf-8'),
                         'gidNumber': uid.encode('utf-8'),
                         'uid': cn.encode('utf-8')}
            user_dn = f'uid={cn},ou=Users,{ds_suffix}'
            (_, _) = ldap_inst.add_entry(user_info, user_dn)

        for (cn, uid, memberUid) in [('childgroup', '111113', 'user2'.encode('utf-8')),
                                     ('middlegroup', '111114', 'childgroup'.encode('utf-8')),
                                     ('topgroup', '111115', ['middlegroup'.encode('utf-8'),
                                                             'user1'.encode('utf-8')])]:
            grop_info = {'cn': cn.encode('utf-8'),
                         'objectClass': b'posixGroup',
                         'gidNumber': uid.encode('utf-8'),
                         'memberUid': memberUid}
            group_dn = f'cn={cn},ou=Groups,{ds_suffix}'
            (_, _) = ldap_inst.add_entry(grop_info, group_dn)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for command in ["id user1 | grep topgroup",
                        "id user2 | grep childgroup",
                        "getent group middlegroup",
                        "getent group topgroup | grep user1",
                        "getent group childgroup | grep user2"]:
            execute_cmd(multihost, command)
        for cn in ['user1',
                   'user2']:
            ldap_inst.del_dn(f'uid={cn},ou=Users,{ds_suffix}')
        for cn in ['childgroup',
                   'middlegroup',
                   'topgroup']:
            ldap_inst.del_dn(f'cn={cn},ou=Groups,{ds_suffix}')

    def test_fully_qualified_names(self, multihost, backupsssdconf):
        """
        :title: fully qualified names
        :id: fd049b9e-4332-11ed-a4a3-845cf3eff344
        :steps:
          1. Look for fully qualified names
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'use_fully_qualified_names': 'true'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for command in ["getent passwd User_CS1@example1",
                        "getent group User_CS1_grp1@example1",
                        "id User_CS1@example1"]:
            execute_cmd(multihost, command)
        check_login_client(multihost, "User_CS1@example1", "Secret123")

    def test_min_id_max_id(self, multihost, backupsssdconf):
        """
        :title: min id max id
        :id: 02ed13ec-4333-11ed-86e7-845cf3eff344
        :steps:
          1. Tests for min id max id
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'min_id': '111112'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for command in ["getent passwd User_CS1",
                        "getent group User_CS1_grp1"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, command)
        execute_cmd(multihost, "sed -i '/min_id/d' /etc/sssd/sssd.conf")
        execute_cmd(multihost, 'echo "max_id=111110" >> /etc/sssd/sssd.conf')
        tools.clear_sssd_cache()
        for command in ["getent passwd User_CS1",
                        "getent group User_CS1_grp1"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, command)
        execute_cmd(multihost, "sed -i '/max_id/d' /etc/sssd/sssd.conf")
        execute_cmd(multihost, 'echo "min_id=111110" >> /etc/sssd/sssd.conf')
        execute_cmd(multihost, 'echo "max_id=111112" >> /etc/sssd/sssd.conf')
        tools.clear_sssd_cache()
        for command in ["getent passwd User_CS1",
                        "getent group User_CS1_grp1",
                        "id User_CS1"]:
            execute_cmd(multihost, command)

    def test_case_sensitive_false_lookup_user_group(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false lookup user and group
        :id: 082124fc-4333-11ed-8b53-845cf3eff344
        :steps:
          1. While case_sensitive=false, look out for user/groups
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        execute_cmd(multihost, 'echo "ignorecase yes" >> /etc/nslcd.conf')
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'false'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        execute_cmd(multihost, 'systemctl restart nslcd')
        tools.clear_sssd_cache()
        for command in ["getent passwd user_cs1 | grep user_cs1",
                        "getent passwd User_CS1 | grep user_cs1",
                        "getent passwd user_cs1_alias | grep user_cs1",
                        "getent passwd User_CS1_Alias | grep user_cs1",
                        "getent group User_CS1_grp1 | grep user_cs1_grp1",
                        "getent group user_cs1_grp1 | grep user_cs1",
                        "getent group User_CS1_grp1_Alias | grep user_cs1_grp1",
                        "getent group user_cs1_grp1_alias | grep user_cs1",
                        "id User_cs1 | grep user_cs1_grp1",
                        "id user_cs1_Alias | grep user_cs1_grp1"]:
            execute_cmd(multihost, command)
        for user in ['user_cs1', 'user_cs1_alias']:
            check_login_client(multihost, user, "Secret123")

    def test_case_sensitive_false_deny_users_user_cs1(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false simple deny user is user_cs1
        :id: 0d8f466c-4333-11ed-9d4d-845cf3eff344
        :steps:
          1. While case_sensitive=false, simple deny user is user_cs1
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'false',
                         'access_provider': 'simple',
                         'simple_deny_users': 'user_cs1'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for user in ['User_cs1', 'user_cs1_alias']:
            execute_cmd(multihost, "> /var/log/secure")
            with pytest.raises(Exception):
                check_login_client(multihost, user, "Secret123")
            time.sleep(3)
            execute_cmd(multihost, f'cat /var/log/secure | grep "Access denied for user {user}"')
        execute_cmd(multihost, "> /var/log/secure")
        execute_cmd(multihost, "sed -i 's/user_cs1/user_cs1_alias/' /etc/sssd/sssd.conf")
        tools.clear_sssd_cache()
        for user in ['user_cs1', 'user_cs1_alias']:
            check_login_client(multihost, user, "Secret123")

    def test_case_sensitive_false_deny_groups_user_cs1_grp1(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false simple deny groups grp1
        :id: 130ccdda-4333-11ed-bdcf-845cf3eff344
        :steps:
          1. While case_sensitive=false, simple deny groups grp1
        :expectedresults:
          1. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'case_sensitive': 'false',
                         'access_provider': 'simple',
                         'simple_deny_groups': 'user_cs1_grp1'
                         }
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for user in ['User_cs1', 'user_cs1_alias']:
            execute_cmd(multihost, "> /var/log/secure")
            with pytest.raises(Exception):
                check_login_client(multihost, user, "Secret123")
            time.sleep(3)
            execute_cmd(multihost, f'cat /var/log/secure | grep -i "Access denied for user {user}"')
        execute_cmd(multihost, "> /var/log/secure")
        execute_cmd(multihost, "sed -i 's/user_cs1_grp1/user_cs1_grp1_alias/' /etc/sssd/sssd.conf")
        tools.clear_sssd_cache()
        for user in ['user_cs1', 'user_cs1_alias']:
            check_login_client(multihost, user, "Secret123")

    def test_outgoing_ldaps(self, multihost, backupsssdconf):
        """
        :title: proxy provider: outgoing ldaps and ldap ports rejected
        :id: 15709f7e-43b6-11ed-9c3f-845cf3eff344
        :steps:
          1. cache_credentials=true
          2. login with ldap user
          3. Block outgoing ldaps
          3. login with ldap user
          4. Unblock outgoing ldaps
          5. login with ldap user
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
        """
        config_sssd(multihost)
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'access_provider': 'simple', 'cache_credentials': "true"}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        check_login_client(multihost, "User_CS1", "Secret123")
        # block master server
        multihost.client[0].run_command('iptables -A OUTPUT -p tcp --match multiport --dport 389,636')
        time.sleep(5)
        check_login_client(multihost, "User_CS1", "Secret123")
        # unblock master server
        multihost.client[0].run_command("iptables -D OUTPUT -p tcp --match multiport --dport 389,636")
        time.sleep(5)
        check_login_client(multihost, "User_CS1", "Secret123")



#
# Module for simulation of utility "getent group -s sss" from coreutils
#
# Copyright (c) 2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from ctypes import (c_int, c_char_p, c_ulong, POINTER, Structure,
                    create_string_buffer)
from sssd_nss import NssReturnCode, SssdNssError, nss_sss_ctypes_loader

GROUP_BUFLEN = 1024


class Group(Structure):
    _fields_ = [("gr_name", c_char_p),
                ("gr_passwd", c_char_p),
                ("gr_gid", c_int),
                ("gr_mem", POINTER(c_char_p))]


def getgrnam_r(name, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_getgrnam_r(const char *name,
                                            struct group *result,
                                            char *buffer,
                                            size_t buflen,
                                            int *errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_getgrnam_r")
    func.restype = c_int
    func.argtypes = [c_char_p, POINTER(Group),
                     c_char_p, c_ulong, POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))

    name = name.encode('utf-8')
    res = func(c_char_p(name), result_p, buffer_p, buflen, errno)

    return (int(res), int(errno[0]), result_p)


def getgrgid_r(gid, result_p, buffer_p, buflen):
    """
    ctypes wrapper for:
        enum nss_status _nss_sss_getgrgid_r(gid_t gid,
                                            struct passwd *result,
                                            char *buffer,
                                            size_t buflen,
                                            int *errnop)
    """
    func = nss_sss_ctypes_loader("_nss_sss_getgrgid_r")
    func.restype = c_int
    func.argtypes = [c_ulong, POINTER(Group),
                     c_char_p, c_ulong, POINTER(c_int)]

    errno = POINTER(c_int)(c_int(0))

    res = func(gid, result_p, buffer_p, buflen, errno)

    return (int(res), int(errno[0]), result_p)


def set_group_dict(res, result_p):
    if res != NssReturnCode.SUCCESS:
        return dict()

    group_dict = dict()
    group_dict['name'] = result_p[0].gr_name.decode('utf-8')
    group_dict['gid'] = result_p[0].gr_gid
    group_dict['mem'] = list()

    i = 0
    while result_p[0].gr_mem[i] is not None:
        grp_name = result_p[0].gr_mem[i].decode('utf-8')
        group_dict['mem'].append(grp_name)
        i = i + 1

    return group_dict


def call_sssd_getgrnam(name):
    """
    A Python wrapper to retrieve a group by name. Returns:
        (res, group_dict)
    if res is NssReturnCode.SUCCESS, then group_dict contains the keys
    corresponding to the C passwd structure fields. Otherwise, the dictionary
    is empty and errno indicates the error code
    """
    result = Group()
    result_p = POINTER(Group)(result)
    buff = create_string_buffer(GROUP_BUFLEN)

    res, errno, result_p = getgrnam_r(name, result_p, buff, GROUP_BUFLEN)
    if errno != 0:
        raise SssdNssError(errno, "getgrnam_r")

    group_dict = set_group_dict(res, result_p)
    return res, group_dict


def call_sssd_getgrgid(gid):
    """
    A Python wrapper to retrieve a group by GID. Returns:
        (res, group_dict)
    if res is NssReturnCode.SUCCESS, then group_dict contains the keys
    corresponding to the C passwd structure fields. Otherwise, the dictionary
    is empty and errno indicates the error code
    """
    result = Group()
    result_p = POINTER(Group)(result)
    buff = create_string_buffer(GROUP_BUFLEN)

    res, errno, result_p = getgrgid_r(gid, result_p, buff, GROUP_BUFLEN)
    if errno != 0:
        raise SssdNssError(errno, "getgrgid_r")

    group_dict = set_group_dict(res, result_p)
    return res, group_dict



""" Automation of sss_cache tests
:requirement: :IDM-SSSD-REQ:: sss_cache improvements
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import pytest
import ldap
from sssd.testlib.common.utils import sssdTools, LdapOperations


@pytest.mark.usefixtures('setup_sssd_krb', 'create_posix_usersgroups')
@pytest.mark.sss_cache
class TestSssCache(object):
    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__reset_cached_timestamps')
    @pytest.mark.tier1_2
    def test_sss_cache_reset(self, multihost, backupsssdconf):
        """
        :title: fix sss_cache to also reset cached timestamp
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1902280
        :customerscenario: True
        :id: c310f1b4-e89b-11eb-84ce-845cf3eff344
        :steps:
            1. Make a change to group entry in LDAP
            2. Run 'sss_cache -E' on clients
            3. Check with 'getent group' on clients to see if correct\
        :expectedresults:
            1. Should succeed
            2. Should succeed
            3. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        domain_name = tools.get_domain_section_name()
        client = sssdTools(multihost.client[0])
        domain_params = {'ldap_schema': 'rfc2307bis',
                         'ldap_group_member': 'uniquemember',
                         'debug_level': '9'}
        client.sssd_conf(f'domain/{domain_name}', domain_params)
        multihost.client[0].service_sssd('restart')
        get_ent = multihost.client[0].run_command("getent group "
                                                  "ldapusers@example1")
        assert "foo9@example1" in get_ent.stdout_text
        user_dn = 'uid=foo9,ou=People,dc=example,dc=test'
        group_dn = 'cn=ldapusers,ou=Groups,dc=example,dc=test'
        ldap_uri = 'ldap://%s' % (multihost.master[0].sys_hostname)
        ds_rootdn = 'cn=Directory Manager'
        ds_rootpw = 'Secret123'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        del_member = [(ldap.MOD_DELETE, 'uniqueMember',
                       user_dn.encode('utf-8'))]
        (ret, _) = ldap_inst.modify_ldap(group_dn, del_member)
        assert ret == 'Success'
        multihost.client[0].run_command("sss_cache -G")
        multihost.client[0].run_command("sss_cache -E")
        get_ent1 = multihost.client[0].run_command("getent group "
                                                   "ldapusers@example1")
        assert "foo9@example1" not in get_ent1.stdout_text
        assert get_ent.stdout_text != get_ent1.stdout_text



"""Automation tests for sudo

:requirement: sudo
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import time
import re
import pytest
from sssd.testlib.common.ssh2_python import SSHClient
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name, ds_suffix


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups',
                         'enable_sss_sudo_nsswitch')
@pytest.mark.sudo
class TestSudo(object):
    """ Sudo test suite """

    @staticmethod
    @pytest.mark.converted('test_sudo.py', 'test_sudo__local_users_negative_cache')
    @pytest.mark.usefixtures('backupsssdconf')
    @pytest.mark.tier1_2
    def test_bz1294670(multihost, localusers):
        """
        :title: sudo: Local users with local sudo rules causes LDAP queries
        :id: e8c5c396-e5e5-4eff-84f8-feff01defda1
        """
        # enable sudo with authselect
        authselect_cmd = 'authselect select sssd with-sudo'
        multihost.client[0].run_command(authselect_cmd)

        # stop sssd service
        multihost.client[0].service_sssd('stop')
        tools = sssdTools(multihost.client[0])
        # remove sssd cache
        tools.remove_sss_cache('/var/lib/sss/db/')
        ldap_uri = 'ldap://%s' % multihost.master[0].sys_hostname
        sssd_params = {'services': 'nss, pam, sudo'}
        tools.sssd_conf('sssd', sssd_params)
        ldap_params = {'ldap_uri': ldap_uri}
        tools.sssd_conf('domain/%s' % (ds_instance_name), ldap_params)
        multihost.client[0].service_sssd('restart')
        sudo_pcapfile = '/tmp/bz1294670.pcap'
        ldap_host = multihost.master[0].sys_hostname
        tcpdump_cmd = 'tcpdump -s0 host %s -w %s' % (ldap_host, sudo_pcapfile)
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        for user in localusers.keys():
            add_rule1 = "echo '%s  ALL=(ALL) NOPASSWD:ALL,!/bin/sh'"\
                        " >> /etc/sudoers.d/%s" % (user, user)
            multihost.client[0].run_command(add_rule1)
            add_rule2 = "echo 'Defaults:%s !requiretty'"\
                        " >> /etc/sudoers.d/%s" % (user, user)
            multihost.client[0].run_command(add_rule2)
            ssh = SSHClient(multihost.client[0].sys_hostname, user, 'Secret123')
            try:
                ssh.connect()
                for _ in range(1, 10):
                    ssh.execute_command('sudo fdisk -l')
                    ssh.execute_command('sudo ls -l /usr/sbin/')
                ssh.close()
            except Exception:
                pytest.fail(f"Authentication Failed as user {user}")
        pkill = 'pkill tcpdump'
        multihost.client[0].run_command(pkill)
        for user in localusers.keys():
            rm_sudo_rule = "rm -f /etc/sudoers.d/%s" % (user)
            multihost.client[0].run_command(rm_sudo_rule)
        tshark_cmd = 'tshark -r %s -R ldap.filter -V -2' % sudo_pcapfile
        cmd = multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
        print("output = ", cmd.stderr_text)
        assert cmd.returncode == 0
        rm_pcap_file = 'rm -f %s' % sudo_pcapfile
        multihost.client[0].run_command(rm_pcap_file)

    @staticmethod
    @pytest.mark.converted('test_sudo.py', 'test_sudo__sudonotbefore_shorttime')
    @pytest.mark.usefixtures('backupsssdconf')
    @pytest.mark.tier2
    def test_timed_sudoers_entry(multihost, timed_sudoers):
        """
        :title: sudo: sssd accepts timed entries without minutes and or
         seconds to attribute
        :id: 5103a796-6c7f-4af0-b7b8-64c7338f0934
        """
        # pylint: disable=unused-argument
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db/')
        sudo_base = 'ou=sudoers,dc=example,dc=test'
        sudo_uri = "ldap://%s" % multihost.master[0].sys_hostname
        params = {'ldap_sudo_search_base': sudo_base,
                  'ldap_uri': sudo_uri, 'sudo_provider': "ldap"}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, params, action='update')
        section = "sssd"
        sssd_params = {'services': 'nss, pam, sudo'}
        tools.sssd_conf(section, sssd_params, action='update')
        multihost.client[0].service_sssd('start')
        ssh = SSHClient(multihost.client[0].sys_hostname, 'foo1@example.test', 'Secret123')
        try:
            ssh.connect()
            id_out = ssh.execute_command('id')
            sudo_out = ssh.execute_command('sudo -l')
            ssh.close()
        except Exception:
            pytest.fail("Failed to login via ssh.")
        assert 'foo1' in id_out, "id command did not work."
        assert 'NOTBEFORE=' in sudo_out or 'NOTAFTER=' in sudo_out,\
            "Expected sudo rule not found!"
        # Make sure that the rule validity time works without minutes
        # and seconds 0000Z is at the end of the NOTAFTER part of rule
        rule_time = re.search(
            r"(NOTBEFORE|NOTAFTER)=[0-9]{10}0000Z NOPASSWD: /usr/bin/head",
            sudo_out)
        if not rule_time:
            journalctl_cmd = 'journalctl -x -n 100 --no-pager'
            multihost.master[0].run_command(journalctl_cmd)
            pytest.fail("sudo -l cmd failed for user foo1")

    @staticmethod
    @pytest.mark.converted('test_sudo.py', 'test_sudo__refresh_random_offset')
    @pytest.mark.usefixtures('backupsssdconf', 'sudo_rule')
    @pytest.mark.tier2
    def test_randomize_sudo_timeout(multihost):
        """
        :title: sudo: randomize sudo refresh timeouts
        :id: 57720975-29ba-4ed7-868a-f9b784bbfed2
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1925514
        :customerscenario: True
        :steps:
          1. Edit sssdconfig and specify sssd smart, full timeout option
          2. Restart sssd with cleared logs and cache
          3. Wait for 120 seconds
          4. Parse logs and confirm sudo refresh timeouts are random
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        tools.remove_sss_cache('/var/log/sssd')
        sudo_base = 'ou=sudoers,%s' % (ds_suffix)
        sudo_uri = "ldap://%s" % multihost.master[0].sys_hostname
        params = {'ldap_sudo_search_base': sudo_base,
                  'ldap_uri': sudo_uri,
                  'sudo_provider': "ldap",
                  'ldap_sudo_full_refresh_interval': '25',
                  'ldap_sudo_smart_refresh_interval': '15',
                  'ldap_sudo_random_offset': '5'}
        domain_section = 'domain/%s' % ds_instance_name
        tools.sssd_conf(domain_section, params, action='update')
        section = "sssd"
        sssd_params = {'services': 'nss, pam, sudo'}
        tools.sssd_conf(section, sssd_params, action='update')
        multihost.client[0].service_sssd('start')
        time.sleep(120)
        logfile = '/var/log/sssd/sssd_%s.log' % ds_instance_name
        tmout_ptrn = r"(SUDO.*\:\sscheduling task \d+ seconds)"
        regex_tmout = re.compile("%s" % tmout_ptrn)
        smart_tmout = []
        full_tmout = []
        log = multihost.client[0].get_file_contents(logfile).decode('utf-8')
        for line in log.split('\n'):
            if line:
                if regex_tmout.findall(line):
                    rfrsh_type = regex_tmout.findall(line)[0].split()[1]
                    timeout = regex_tmout.findall(line)[0].split()[5]
                    if rfrsh_type == 'Smart':
                        smart_tmout.append(timeout)
                    elif rfrsh_type == 'Full':
                        full_tmout.append(timeout)
        rand_intvl, same_intvl = 0, 0
        for timeout in smart_tmout, full_tmout:
            index = 1
            rand_intvl, same_intvl = 0, 0
            while index < len(timeout):
                if timeout[index] != timeout[index - 1]:
                    rand_intvl += 1
                else:
                    same_intvl += 1
                index += 1
            assert rand_intvl > same_intvl

    @staticmethod
    @pytest.mark.converted('test_sudo.py', 'test_sudo__prefer_full_refresh_over_smart_refresh')
    @pytest.mark.usefixtures('backupsssdconf', 'sudo_rule', 'sssd_sudo_conf')
    @pytest.mark.tier2
    def test_improve_refresh_timers_sudo_timeout(multihost):
        """
        :title: sudo: improve sudo full and smart refresh timeouts
        :id: 3860d1b9-28fc-4d44-9537-caf28ab033c8
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1925505
        :customerscenario: True
        :steps:
          1. Edit sssdconfig and specify sssd smart, full timeout option
          2. Restart sssd with cleared logs and cache
          3. Wait for 40 seconds
          4. Parse logs and confirm sudo full refresh and smart refresh
             timeout are not running at same time
          5. If sudo full refresh and smart refresh timer are scheduled at
             same time then smart refresh is rescheduled to the next cycle
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        tools.remove_sss_cache('/var/log/sssd')
        params = {'ldap_sudo_full_refresh_interval': '10',
                  'ldap_sudo_random_offset': '0',
                  'ldap_sudo_smart_refresh_interval': '5'}
        domain_section = f'domain/{ds_instance_name}'
        tools.sssd_conf(domain_section, params, action='update')
        multihost.client[0].service_sssd('start')
        time.sleep(40)
        logfile = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        tmout_ptrn = '(SUDO.*Refresh.*executing)'
        rschdl_ptrn = '(SUDO.*Refresh.*rescheduling)'
        regex_tmout = re.compile(f'{tmout_ptrn}')
        rgx_rs_tstmp = re.compile(f'{rschdl_ptrn}')
        full_rfsh_tstmp = []
        smrt_rfsh_tstmp = []
        rschdl_tstmp = []
        log = multihost.client[0].get_file_contents(logfile).decode('utf-8')
        for line in log.split('\n'):
            if regex_tmout.findall(line):
                dt_time = line.split('):')[0]
                tstmp = dt_time.split()[1]
                ref_type = line.split()[7]
                if ref_type == 'Smart':
                    smrt_rfsh_tstmp.append(tstmp)
                elif ref_type == 'Full':
                    full_rfsh_tstmp.append(tstmp)
            if rgx_rs_tstmp.findall(line):
                dt_time = line.split('):')[0]
                tstmp = dt_time.split()[1]
                rschdl_tstmp.append(tstmp)
        for tm_stamp in full_rfsh_tstmp:
            if tm_stamp in smrt_rfsh_tstmp:
                assert tm_stamp in rschdl_tstmp
            else:
                assert tm_stamp not in smrt_rfsh_tstmp
